"""This module allows getting information and changing Alma users"""

from typing import Optional, ClassVar, Literal, Union, List
import logging

import almapiwrapper.inventory

from ..record import Record, check_error, JsonData
import almapiwrapper.users as users
from datetime import datetime


class User(Record):
    """Class representing an Alma user

    :ivar primary_id: initial value: primary_id of the user
    :ivar zone: initial value: zone where the user should be created
    :ivar env: initial value: environment of the entity: 'P' for production and 'S' for sandbox
    :ivar data: initial value: :class:`almapiwrapper.record.JsonData` of the user (useful for new created users)
    """

    api_base_url: ClassVar[str] = 'https://api-eu.hosted.exlibrisgroup.com/almaws/v1/users'

    def __init__(self,
                 primary_id: str,
                 zone: str,
                 env: Optional[Literal['P', 'S']] = 'P',
                 data: Optional[JsonData] = None) -> None:
        """Constructor for user
        """
        super().__init__(zone, env, data)
        self.primary_id = primary_id
        self.area = 'Users'
        self.format = 'json'
        self._fees = None
        self._loans = None
        self._requests = None

    def __repr__(self) -> str:
        """Get a string representation of the object. Useful for logs.
        :return: string
        """
        return f"{self.__class__.__name__}('{self.primary_id}', '{self.zone}', '{self.env}')"

    def _fetch_data(self) -> Optional[JsonData]:
        """Use API to fetch user data

        :return: json data or None if no data is available
        """

        r = self._api_call('get',
                           f'{self.api_base_url}/{self.primary_id}',
                           headers=self._get_headers())
        if r.ok is True:
            logging.info(f'{repr(self)}: user data available')
            return JsonData(r.json())
        else:
            self._handle_error(r, f'unable to fetch user data')

    def _fetch_fees(self) -> Optional[List['users.Fee']]:
        """Fetch fee data of the current user

        :return: list of :class:`almapiwrapper.users.Fee` objects"""

        r = self._api_call('get',
                           f'{self.api_base_url}/{self.primary_id}/fees',
                           # params={'status': 'EXPORTED'},
                           headers=self._get_headers())
        if r.ok is True:

            logging.info(f'{repr(self)}: fees data available')
            fees_data = r.json()
            if 'fee' not in fees_data or fees_data['fee'] is None:
                logging.warning(f'{repr(self)}: no fee in the account')
                return []
            fees = []
            for fee_data in fees_data['fee']:
                fees.append(users.Fee(user=self, data=JsonData(fee_data)))
            return fees
        else:
            self._handle_error(r, f'unable to fetch user fees')

    def _fetch_loans(self) -> Optional[List['users.Loan']]:
        """Fetch loan data of the current user

        :return: list of :class:`almapiwrapper.users.Loan` objects"""

        r = self._api_call('get',
                           f'{self.api_base_url}/{self.primary_id}/loans',
                           params={'limit': '100'},
                           headers=self._get_headers())
        if r.ok is True:

            logging.info(f'{repr(self)}: loans data available')
            loans_data = r.json()
            if 'item_loan' not in loans_data or loans_data['item_loan'] is None:
                logging.warning(f'{repr(self)}: no loan in the account')
                return []
            loans = []
            for loan_data in loans_data['item_loan']:
                loans.append(users.Loan(user=self, data=JsonData(loan_data)))
            return loans
        else:
            self._handle_error(r, f'unable to fetch user loans')

    def _fetch_requests(self) -> Optional[List['users.Request']]:
        """Fetch request data of the current user

        :return: list of :class:`almapiwrapper.users.Request` objects"""

        r = self._api_call('get',
                           f'{self.api_base_url}/{self.primary_id}/requests',
                           params={'limit': '100'},
                           headers=self._get_headers())
        if r.ok is True:

            logging.info(f'{repr(self)}: requests data available')
            request_data = r.json()
            if 'user_request' not in request_data or request_data['user_request'] is None:
                logging.warning(f'{repr(self)}: no request in the account')
                return []
            user_requests = []
            for request_data in request_data['user_request']:
                user_requests.append(users.Request(user=self, data=JsonData(request_data)))
            return user_requests
        else:
            self._handle_error(r, f'unable to fetch user requests')

    def save(self) -> 'User':
        """save() -> 'User'
        Save a user record in the 'records' folder

        When saved, a suffix is added to the file path with the version.
        Example: records/<primary_id>/user_<IZ>_<primary_id>_<version>.xml

        :return: object :class:`almapiwrapper.users.User`
        """
        filepath = f'records/{self.primary_id}/user_{self.zone}_{self.primary_id}.json'
        self._save_from_path(filepath)
        return self

    @check_error
    def update(self, override: Optional[str] = None) -> 'User':
        """update() -> 'User'
        Update the user through api

        :param override: string containing the list of the fields to override

        :return: object :class:`almapiwrapper.users.User`

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """
        if override is not None:
            params = {'override': override}
        else:
            params = {}

        r = self._api_call('put',
                           f'{self.api_base_url}/{self.primary_id}',
                           data=bytes(self),
                           params=params,
                           headers=self._get_headers())
        if r.ok:
            logging.info(f'{repr(self)}: user updated.')
            self.data = JsonData(r.json())
        else:
            self._handle_error(r, 'failed to update user')

        return self

    @check_error
    def delete(self) -> None:
        """delete()
        Delete the user

        :return: None

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """
        r = self._api_call('delete',
                           f'{self.api_base_url}/{self.primary_id}',
                           headers=self._get_headers())
        if r.ok:
            logging.info(f'{repr(self)}: user deleted.')
        else:
            self._handle_error(r, 'failed to delete user')

    @property
    def primary_id(self) -> str:
        """Property returning the primary ID

        It fetches it in a private attribute if not available
        in the data property.

        When a new value is set and json data is available, it changes the primary ID
        at the data level.

        :return: primary ID of the user
        """

        return self._primary_id

    @primary_id.setter
    def primary_id(self, primary_id) -> None:
        """Set the primary ID

        If json data is available, change the primary ID
        at the data level. To modify the primary ID of the user,
        do not use this method property but change directly the primary_id
        in the data attribute.

        :param primary_id: new primary ID to set

        :return: None
        """
        self._primary_id = primary_id
        if self._data is not None:
            self.data['primary_id'] = self._primary_id

    @property
    def fees(self) -> Optional[List['users.Fee']]:
        """Property returning the list of the fees of the user

        :return: list of :class:`almapiwrapper.users.Fee` objects"""
        if self._fees is None:
            self._fees = self._fetch_fees()

        return self._fees

    @property
    def loans(self) -> Optional[List['users.Loan']]:
        """Property returning the list of the loans of the user

        :return: list of :class:`almapiwrapper.users.Loan` objects"""
        if self._loans is None:
            self._loans = self._fetch_loans()

        return self._loans

    @property
    def requests(self) -> Optional[List['users.Request']]:
        """Property returning the list of the requests of the user

        :return: list of :class:`almapiwrapper.users.Request` objects"""
        if self._requests is None:
            self._requests = self._fetch_requests()

        return self._requests

    @check_error
    def create_loan(self,
                    library: str,
                    circ_desk: str,
                    item_barcode: Optional[str] = None,
                    item_id: Optional[str] = None,
                    item: Optional[almapiwrapper.inventory.Item] = None) -> Optional['users.Loan']:
        """create_loan(library: str, circ_desk: str, item_barcode: Optional[str] = None, item_id: Optional[str] = None, item: Optional[almapiwrapper.inventory.Item] = None) -> Optional['users.Loan']

        Create a loan for the user

        :param library: code of the library
        :param circ_desk: code of the circulation desk
        :param item_barcode: barcode of the item
        :param item_id: id of the item
        :param item: object :class:`almapiwrapper.inventory.Item`

        :return: object :class:`almapiwrapper.users.Loan`

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """
        if item is not None and item.error is False:
            item_barcode = item.barcode

        # Check if mandatory fields are available
        if library is None or circ_desk is None or (item_barcode is None and item_id is None):
            logging.error('Missing information to create a loan')
            self.error = True
            return None

        # Create the loan
        loan_data = {"circ_desk": {"value": circ_desk},
                     "library": {"value": library}}

        params = {'item_barcode': item_barcode} if item_barcode is not None else {'item_pid': item_id}

        r = self._api_call('post',
                           f'{self.api_base_url}/{self.primary_id}/loans',
                           params=params,
                           data=bytes(JsonData(loan_data)),
                           headers=self._get_headers())
        if r.ok is True:
            loan_data = JsonData(r.json())

            loan = users.Loan(user=self, data=loan_data)

            logging.info(f'{repr(self)}: {repr(loan)} created.')

            return loan
        else:
            self._handle_error(r, f'unable to create loan for '
                                  f'{item_barcode if item_barcode is not None else item_id} '
                                  f'on {circ_desk} at {library}')

    @check_error
    def set_password(self, password: Optional[str] = '123pw123') -> 'User':
        """set_password(password: Optional[str] = '123pw123') -> 'User'
        Set a new password to a user

        :param password: password string, default is '123pw123'

        :return: object :class:`almapiwrapper.users.User`

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """
        self.data['password'] = password
        self.data['force_password_change'] = "TRUE"

        return self

    @check_error
    def add_synchro_note(self) -> 'User':
        """add_synchro_note() -> 'User'
        Add a test synchronization notes

        :return: object :class:`almapiwrapper.users.User`"""

        note = {"note_type": {"value": "OTHER"},
                "note_text": f"SLSP Test-synchronisation {datetime.now().strftime('%Y-%m-%d %H:%M')}",
                "user_viewable": False,
                "popup_note": False,
                "segment_type": "Internal"}
        self.data['user_note'].append(note)
        self.update()
        if self.error is False:
            logging.info(f'{repr(self)}: synchronization test note added')

        return self

    @check_error
    def remove_synchro_note(self) -> 'User':
        """remove_synchro_note() -> 'User'
        Remove all test synchronization notes

        :return: object :class:`almapiwrapper.users.User`"""
        nb_notes = len(self.data['user_note'])
        self.data['user_note'] = [note for note in self.data['user_note']
                                  if not note['note_text'].startswith('SLSP Test-synchronisation')]
        nb_sup_notes = nb_notes - len(self.data['user_note'])

        if nb_sup_notes > 0:
            self.update()
            if self.error is False:
                logging.info(f'{repr(self)}: {nb_sup_notes} synchronization test note(s) deleted')
        else:
            logging.warning(f'{repr(self)}: NO synchronization test note deleted')

        return self

    @check_error
    def check_synchro_note(self) -> bool:
        """check_synchro_note() -> bool
        Test if the user has a synchronization test note

        :return: True if a synchronization test note is found else False
        """
        has_synchro_note = len([note for note in self.data['user_note']
                                if note['note_text'].startswith('SLSP Test-synchronisation')]) > 0
        if has_synchro_note is True:
            logging.info(f'{repr(self)}: a synchronization test note exists')
        else:
            logging.warning(f'{repr(self)}: NO synchronization test note exists')
        return has_synchro_note

    @check_error
    def dedup_notes(self) -> 'User':
        """dedup_notes(self) -> 'User'
        Remove all duplicated notes

        Text of note is used to compare notes.
        If the text is the same, the external note is preferred.

        :return: object :class:`almapiwrapper.users.User`
        """
        notes_to_be_kept = []
        for note in self.data['user_note']:
            found = False
            for note_to_be_kept in notes_to_be_kept:
                if note_to_be_kept['note_text'] == note['note_text']:
                    found = True

                    # Prefer external note
                    if note['segment_type'] == 'External' and note_to_be_kept['segment_type'] == 'Internal':
                        # remove internal note and add the external one to be kept
                        notes_to_be_kept.remove(note_to_be_kept)
                        notes_to_be_kept.append(note)
                    break

            # Add the current tested note if no corresponding note is found
            if found is False:
                notes_to_be_kept.append(note)
        self.data['user_note'] = notes_to_be_kept
        return self

    def fetch_requests(self) -> Optional[List['users.Request']]:
        """fetch_requests(self) -> Optional[List['users.Request']]
        Fetch the list of the requests of the user"""

        pass

class NewUser(User):
    """Class used to create new users

    :var zone: initial value: zone where the user should be created
    :var env: initial value: environment of the entity: 'P' for production and 'S' for sandbox
    :var data: initial value: :class:`almapiwrapper.record.JsonData` of the user to create
    """

    def __init__(self, zone: str, env: Literal['P', 'S'], data: JsonData):
        """Constructor to create new users
        """
        super(User, self).__init__(zone, env, data)
        self._primary_id = self.data['primary_id']
        self.area = 'Users'
        self.format = 'json'

    @check_error
    def create(self, password: Optional[str] = None) -> Union['NewUser', User]:
        """create(password: Optional[str] = None) -> Union['NewUser', User]
        Create the user with API.

        :param password: optional string with the password, if not provided,
            the password will be set at the default value (if not available in the data)

        :return: object :class:`almapiwrapper.users.User` or object :class:`almapiwrapper.users.NewUser`
            (in case of error)

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """
        self.set_password(password)

        r = self._api_call('post',
                           f'{self.api_base_url}',
                           headers=self._get_headers(),
                           data=bytes(self))

        if r.ok is True:
            logging.info(f'{repr(self)}: user created')
            self._data = JsonData(r.json())
            return User(self.primary_id, self.zone, self.env, self._data)
        else:
            self._handle_error(r, f'unable to create user')
            return self

    @check_error
    def set_password(self, password: Optional[str] = None) -> 'NewUser':
        """set_password(password: Optional[str] = None) -> 'NewUser'
        Set the password of a new user

        :param password: string containing the new password.

        :return: object :class:`almapiwrapper.users.NewUser`
        .. note::
            If the record encountered an error, this
            method will be skipped.
        """
        if password is None and self.data['password'] == '':
            password = '123pw123'

        if password is not None:
            self.data['password'] = password

        return self


if __name__ == "__main__":
    pass



"""This module allow to create and manage sets in Alma"""
from typing import Literal, Optional, List, Dict, Union
import logging
from ..record import JsonData, Record, check_error
import requests


def fetch_libraries(zone: str,
                    env: Optional[Literal['P', 'S']] = 'P') -> List['Library']:
    """This function fetch the data describing the libraries.

    :param zone: institutional zone
    :param env: environment of the entity: 'P' for production and 'S' for sandbox

    :return: List of :class:`almapiwrapper.config.library.Library`
    """

    libraries = []

    r = requests.get(f'{Library.api_base_url}/conf/libraries',
                     headers=Record.build_headers(data_format='json', env=env,
                                                  zone=zone, rights='RW', area='Conf'))

    # Check result
    if r.ok is True and 'library' in r.json():
        libraries_list = JsonData(r.json())
        libraries = [Library(code=lib_data['code'], zone=zone, env=env, data=lib_data)
                     for lib_data in libraries_list.content['library']]
    elif r.ok is True:
        _handle_error(r, 'no libraries data available', zone, env)

    else:
        _handle_error(r, 'unable to fetch libraries data', zone, env)

    return libraries

def _handle_error(r: requests.models.Response, msg: str, zone: str, env: Literal['P', 'S']) -> None:
    """Set the record error attribute to True and write the logs about the error

    :param r: request response of the api
    :param msg: context message of the error
    :return: None
    """
    json_data = r.json()
    error_message = json_data['errorList']['error'][0]['errorMessage']

    logging.error(f'fetch_reminders({zone}, {env}) - {r.status_code}: '
                  f'{msg} / {error_message}')


class Library(Record):
    """Class representing a reminder
    """

    def __init__(self,
                 code: Optional[str] = None,
                 zone: Optional[str] = None,
                 env: Optional[Literal['P', 'S']] = 'P',
                 library_id: Optional[str] = None,
                 data: Optional[Union[Dict, JsonData]] = None) -> None:
        """Constructor of `Reminder`

        :param code: code of the library
        :param zone: zone of the library
        :param env: environment of the library: 'P' for production and 'S' for sandbox
        :param library_id: library id
        :param data: data of the library, :class:`almapiwrapper.record.JsonData` object

        """
        super().__init__(zone, env, data)
        self.area = 'Conf'
        self.format = 'json'
        self.library_id = library_id
        self.code = code
        self._locations = None
        self._desks = None
        self._open_hours = None

        if data is not None:
            if not isinstance(data, JsonData):
                data = JsonData(data)
            self.data = data


    def _fetch_data(self) -> Optional[JsonData]:
        """This method fetch the data describing the reminder.

        :return: :class:`almapiwrapper.record.JsonData` object
        """
        r = self._api_call('get',
                           f'{self.api_base_url}/conf/libraries/{self.code}',
                           headers=self._get_headers())
        if r.ok is True:
            logging.info(f'{repr(self)}: library data available')
            return JsonData(r.json())
        else:
            self._handle_error(r, 'unable to fetch library data')

    def __repr__(self) -> str:
        """Get a string representation of the object. Useful for logs.

        :return: string
        """
        return f"{self.__class__.__name__}('{self.code}', '{self.zone}', '{self.env}')"

    @check_error
    def _fetch_locations(self) -> List['Location']:
        """Return the list of locations

        :return: list of locations
        """
        r = self._api_call('get',
                           f'{self.api_base_url}/conf/libraries/{self.code}/locations',
                           headers=self._get_headers())
        if r.ok is True and 'location' in r.json():
            logging.info(f'{repr(self)}: locations data available')
            return [Location(library_code=self.code, zone=self.zone, env=self.env, data=location_data) for location_data in r.json()['location']]
        elif r.ok and 'location' not in r.json():
            logging.warning(f'{repr(self)}: no locations data available')
            return []
        else:
            self._handle_error(r, 'unable to fetch locations data')

    @property
    def locations(self) -> List['Location']:
        """Property to get the locations of the library

        :return: list of locations of the library
        """
        if self._locations is None:
            self._locations = self._fetch_locations()

        return self._locations


    @check_error
    def _fetch_desks(self) -> List['Desk']:
        """get_locations(self) -> List
        Return the list of locations

        :return: list of locations
        """
        r = self._api_call('get',
                           f'{self.api_base_url}/conf/libraries/{self.code}/circ-desks',
                           headers=self._get_headers())
        if r.ok is True and 'circ_desk' in r.json():
            logging.info(f'{repr(self)}: circulation desks data available')
            return [Desk(library_code=self.code, code=desk_data['code'], zone=self.zone, env=self.env) for desk_data in r.json()['circ_desk']]
        elif r.ok and 'circ_desk' not in r.json():
            logging.warning(f'{repr(self)}: no circulation desks data available')
            return []
        else:
            self._handle_error(r, 'unable to fetch circulation desks data')

    @property
    def desks(self) -> List['Desk']:
        """Property to get the desks of the library

        :return: list of desks of the library
        """
        if self._desks is None:
            self._desks = self._fetch_desks()

        return self._desks

    @property
    def open_hours(self) -> 'OpenHours':
        """open_hours(self) -> 'OpenHours'
        Get the open hours of the library

        :return: object :class:`almapiwrapper.library.OpenHours`
        """
        if self._open_hours is None:
            self._open_hours = OpenHours(self.zone, self.code, self.env)
        return self._open_hours

    @open_hours.setter
    def open_hours(self, open_hours: 'OpenHours') -> None:
        """open_hours(self, open_hours: 'OpenHours') -> None
        Set the open hours of the library

        :param open_hours: object :class:`almapiwrapper.library.OpenHours`
        :return: None
        """
        self._open_hours = open_hours


class Location(Record):
    """Class representing a location
    """

    def __init__(self,
                 zone: str,
                 library_code: str,
                 code: Optional[str] = None,
                 env: Optional[Literal['P', 'S']] = 'P',
                 data: Optional[Union[Dict, JsonData]] = None) -> None:
        """Constructor of `Location`

        :param code: code of the location
        :param zone: zone of the location
        :param env: environment of the location: 'P' for production and 'S' for sandbox
        :param data: data of the location, :class:`almapiwrapper.record.JsonData` object

        """
        super().__init__(zone, env, data)
        self.library_code = library_code
        self.area = 'Conf'
        self.format = 'json'
        self.code = code

        if data is not None:
            if not isinstance(data, JsonData):
                data = JsonData(data)
            self.code = data.content['code']
            self.data = data

    def __repr__(self) -> str:
        """Get a string representation of the object. Useful for logs.

        :return: string
        """
        return f"{self.__class__.__name__}('{self.zone}', '{self.library_code}', '{self.code}', '{self.env}')"


    def _fetch_data(self) -> Optional[JsonData]:
        """This method fetch the data describing the location.

        :return: :class:`almapiwrapper.record.JsonData` object
        """
        r = self._api_call('get',
                           f'{self.api_base_url}/conf/libraries/{self.library_code}/locations/{self.code}',
                           headers=self._get_headers())
        if r.ok is True:
            logging.info(f'{repr(self)}: location data available')
            return JsonData(r.json())
        else:
            self._handle_error(r, 'unable to fetch location data')

    @check_error
    def save(self) -> 'Location':
        """save() -> 'Location'
        Save a Location record in the 'records' folder

        When saved, a suffix is added to the file path with the version.
        Example: records/<IZ>_<library_code>/location_<IZ>_<library_code>_<location_code>_<version>.xml

        :return: object :class:`almapiwrapper.library.Location`
        """
        filepath = f'records/{self.zone}_{self.library_code}/location_{self.zone}_{self.library_code}_{self.code}.json'
        self._save_from_path(filepath)
        return self

    @check_error
    def update(self) -> 'Location':
        """update() -> Location
        Update the location
        """
        r = self._api_call('put',
                           f'{self.api_base_url}/conf/libraries/{self.library_code}/locations/{self.code}',
                           headers=self._get_headers(),
                           data=bytes(self))

        if r.ok:
            logging.info(f'{repr(self)}: location updated.')
        else:
            self._handle_error(r, 'failed to update location')

        return self

    @check_error
    def create(self) -> 'Location':
        """create() -> 'Location'
        Create a new location

        :return: object :class:`almapiwrapper.library.Location`
        """
        r = self._api_call('post',
                           f'{self.api_base_url}/conf/libraries/{self.library_code}/locations',
                           headers=self._get_headers(),
                           data=bytes(self))

        if r.ok is True:
            logging.info(f'{repr(self)}: new location created')
        else:
            self._handle_error(r, 'unable to create new location')

        return self

    @check_error
    def delete(self) -> None:
        """create() -> None
        Delete a location
        """
        r = self._api_call('delete',
                           f'{self.api_base_url}/conf/libraries/{self.library_code}/locations/{self.code}',
                           headers=self._get_headers())

        if r.ok is True:
            logging.info(f'{repr(self)}: location deleted')
        else:
            self._handle_error(r, 'unable to create new location')

        return None

    @property
    def fulfillment_unit(self) -> str:
        """fulfillment_unit() -> str
        Get the fulfillment unit

        :return: string containing the fulfillment unit
        """
        return self.data['fulfillment_unit']['value']

    @fulfillment_unit.setter
    def fulfillment_unit(self, fulfillment_unit: str) -> None:
        """fulfillment_unit(self, fulfillment_unit: str) -> None
        Set the fulfillment unit

        :return: None
        """
        self.data['fulfillment_unit']['value'] = fulfillment_unit
        if 'desc' in self.data['fulfillment_unit']:
            del self.data['fulfillment_unit']['desc']


class OpenHours(Record):
    """Class representing list of open hours
    """
    def __init__(self,
                 zone: str,
                 library_code: Optional[str] = None,
                 env: Optional[Literal['P', 'S']] = 'P',
                 data: Optional[Union[Dict, JsonData]] = None) -> None:
        """Constructor of `OpenHours`

        :param zone: zone of the open hours
        :param library_code: code of the library
        :param env: environment of the open hours: 'P' for production and 'S' for sandbox
        :param data: data of the open hours, :class:`almapiwrapper.record.JsonData` object
        """

        super().__init__(zone, env, data)
        self.library_code = library_code
        self.area = 'Conf'
        self.format = 'json'

        if data is not None:
            if not isinstance(data, JsonData):
                data = JsonData(data)
            self.data = data

    def __repr__(self) -> str:
        """Get a string representation of the object. Useful for logs.

        :return: string
        """
        if self.library_code is not None:
            return f"{self.__class__.__name__}('{self.zone}', '{self.library_code}', '{self.env}')"
        else:
            return f"{self.__class__.__name__}('{self.zone}', '{self.env}')"

    def _fetch_data(self) -> Optional[JsonData]:
        """This method fetch the data describing the open hours.

        :return: :class:`almapiwrapper.record.JsonData` object
        """
        if self.library_code is not None:
            params = {'scope': self.library_code}
        else:
            params = {}
        r = self._api_call('get',
                           f'{self.api_base_url}/conf/open-hours',
                           params=params,
                           headers=self._get_headers())
        if r.ok is True:
            logging.info(f'{repr(self)}: open hours data available')
            return JsonData(r.json())
        else:
            self._handle_error(r, 'unable to fetch open hours data')

    @check_error
    def update(self) -> 'OpenHours':
        """update(self) -> 'OpenHours'
        Update the open hours

        :return: object :class:`almapiwrapper.library.OpenHours`
        """
        r = self._api_call('put',
                           f'{self.api_base_url}/conf/open-hours',
                           headers=self._get_headers(),
                           data=bytes(self))

        if r.ok:
            logging.info(f'{repr(self)}: open hours updated.')
        else:
            self._handle_error(r, 'failed to update open hours')

        return self

    @check_error
    def save(self) -> 'OpenHours':
        """save(self) -> 'OpenHours'
        Save a OpenHours record in the 'records' folder

        When saved, a suffix is added to the file path with the version.
        Example: records/<IZ>/open_hours_<IZ>_<version>.xml

        :return: object :class:`almapiwrapper.library.OpenHours`
        """
        if self.library_code is not None:
            filepath = f'records/{self.zone}/open_hours_{self.zone}_{self.library_code}.json'
        else:
            filepath = f'records/{self.zone}/open_hours_{self.zone}.json'
        self._save_from_path(filepath)
        return self


class Desk(Record):
    """Class representing a desk

    :ivar zone: zone of the desk
    :ivar library_code: code of the library
    :ivar code: code of the desk
    :ivar env: environment of the desk: 'P' for production and 'S' for sandbox
    :ivar data: data of the desk, :class:`almapiwrapper.record.JsonData` object

    """
    def __init__(self,
                 zone: str,
                 library_code: Optional[str] = None,
                 code: Optional[str] = None,
                 env: Optional[Literal['P', 'S']] = 'P') -> None:
        """Constructor of `Desk`

        :param zone: zone of the desk
        :param library_code: code of the library
        :param code: code of the desk
        :param env: environment of the desk: 'P' for production and 'S' for sandbox
        """
        super().__init__(zone, env)
        self.library_code = library_code
        self.area = 'Conf'
        self.format = 'json'
        self.code = code

    def __repr__(self) -> str:
        """Get a string representation of the object. Useful for logs.

        :return: string
        """
        return f"{self.__class__.__name__}('{self.zone}', '{self.library_code}', '{self.code}', '{self.env}')"

    def _fetch_data(self) -> Optional[JsonData]:
        """This method fetch the data describing the desk.

        :return: :class:`almapiwrapper.record.JsonData` object
        """
        r = self._api_call('get',
                           f'{self.api_base_url}/conf/libraries/{self.library_code}/circ-desks/{self.code}',
                           headers=self._get_headers())
        if r.ok is True:
            logging.info(f'{repr(self)}: desk data available')
            return JsonData(r.json())
        else:
            self._handle_error(r, 'unable to fetch desk data')

    @check_error
    def get_locations(self) -> List[Location]:
        """get_locations(self) -> List[Location]
        Return the list of locations

        :return: list of :class:`almapiwrapper.library.Location`
        """

        locations = [Location(self.zone, self.library_code, location['location_code'], self.env)
                     for location in self.data['location']]
        return locations

    @check_error
    def save(self) -> 'Desk':
        """save() -> 'Desk'
        Save a Desk record in the 'records' folder
        """
        filepath = f'records/{self.zone}/desk_{self.zone}_{self.library_code}_{self.code}.json'
        self._save_from_path(filepath)
        return self



from ..record import Record, check_error, JsonData
from typing import Optional, Literal,ClassVar, Union
import almapiwrapper.acquisitions as acquisitionslib
from lxml import etree
import pandas as pd
import logging

class Vendor(Record):
    """Class representing a Vendor

    :ivar vendor_code: initial value: code of the vendor
    :ivar zone: initial value: zone of the vendor
    :ivar env: initial value: environment of the entity: 'P' for production and 'S' for sandbox
    :ivar data: :class:`almapiwrapper.record.JsonData` with raw vendor data
    """

    api_base_url_vendors: ClassVar[str] = f'{Record.api_base_url}/acq/vendors'

    def __init__(self,
                 vendor_code: Optional[str] = None,
                 zone: Optional[str] = None,
                 env: Literal['P', 'S'] = 'P',
                 data: Optional[Union[dict, JsonData]] = None) -> None:
        """Constructor of Vendor Object
        """

        super().__init__(zone, env)
        self.area = 'Acquisitions'
        self.format = 'json'
        self.vendor_code = vendor_code
        if data is not None:
            self._data = data if type(data) is JsonData else JsonData(data)
        elif vendor_code is not None:
            pass
        else:
            self.error = True
            logging.error('Missing information to construct an Vendor')

    def __repr__(self) -> str:
        """Get a string representation of the object. Useful for logs.

        :return: str representing the object
        """
        return f"{self.__class__.__name__}('{self.vendor_code}', '{self.zone}', '{self.env}')"

    def _fetch_data(self) -> Optional[JsonData]:
        """Fetch the json data of the vendor

        :return: :class:`almapiwrapper.record.JsonData` if no error else None
        """

        r = self._api_call('get',
                     f'{self.api_base_url_vendors}/{self.vendor_code}',
                           headers=self._get_headers())
        if r.ok is True:
            # Parse data
            json_data = JsonData(r.json())
            logging.debug(f"{self.__class__.__name__} data fetched")

            return json_data
        else:
            self._handle_error(r, 'unable to fetch vendor data')

    @check_error
    def update(self) -> 'acquisitionslib.Vendor':
        """update(self) -> 'acquisitionslib.Vendor'
        Update the Vendor

        .. note::
            If the record encountered an error, this
            method will be skipped.

        :return: object :class:`almapiwrapper.acquisitions.Vendor`
        """

        r = self._api_call('put',
                           f'{self.api_base_url_vendors}/{self.vendor_code}',
                           headers=self._get_headers(),
                           data=bytes(self))

        if r.ok is True:
            self.data = JsonData(r.json())
            logging.info(f'{repr(self)}: Vendor data updated')
        else:
            self._handle_error(r, 'unable to update vendor data')

        return self

    @check_error
    def save(self) -> 'acquisitionslib.Vendor':
        """save(self) -> acquisitionslib.Vendor
        Save a PO Line record in the 'records' folder

        When saved, a suffix is added to the file path with the version.
        Example: records/vendors/vendor_<IZ>_<set_id>_<version>.xml

        :return: object :class:`almapiwrapper.acquisitions.Vendor`

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """
        filepath = f'records/vendors/vendor_{self.zone}_{self.vendor_code}.json'
        self._save_from_path(filepath)
        return self

    @check_error
    def create(self):
        """create(self) -> acquisitionslib.Vendor
        Create a Vendor

        :note: The vendor code must be unique, same for the account_id of each account.

        :return: object :class:`almapiwrapper.acquisitions.Vendor`
        """
        r = self._api_call('post',
                           f'{self.api_base_url_vendors}',
                           headers=self._get_headers(),
                           data=bytes(self))

        if r.ok is True:
            self._data = JsonData(r.json())
            self.vendor_code = self.data['code']
            logging.info(f'{repr(self)}: Vendor created: {self.vendor_code}')
        else:
            self._handle_error(r, 'unable to create Vendor')

        return self

    @check_error
    def delete(self) -> None:
        """delete(self) -> None"""
        r = self._api_call('delete',
                           f'{self.api_base_url_vendors}/{self.vendor_code}',
                           headers=self._get_headers())
        if r.ok is True:
            logging.info(f'{repr(self)}: Vendor deleted: {self.vendor_code}')
        else:
            self._handle_error(r, 'unable to delete Vendor')



import unittest
import time
from almapiwrapper.users import User, NewUser, fetch_users
from almapiwrapper.record import JsonData
from almapiwrapper import config_log
import os

config_log("test.log")

if os.getcwd().endswith('test'):
    os.chdir('..')

class TestFetchUsers(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        u = User('TestUser4', 'UBS', 'S')
        if u.data is not None:
            u.delete()

    def test_find_user4(self):

        # Create new user
        data = JsonData(filepath='test/data/user_test4.json')
        u1 = NewUser('UBS', 'S', data).create()

        self.assertFalse(u1.error, 'Error in record after creating the user')
        time.sleep(3)
        users = fetch_users('primary_id~TestUser4', 'UBS', 'S')

        self.assertEqual(len(users), 1, "Nombre d'utilisateurs trouvés différent de 1")

        # Delete user
        u1.delete()

    def test_find_not_existing_user(self):

        users = fetch_users('primary_id~testUser_not_existing', 'UBS', 'S')
        self.assertEqual(len(users), 0, "Nombre d'utilisateurs trouvés différent de 0"
                                        "pour un utilisateur non existant")

    @classmethod
    def tearDownClass(cls):
        u = User('TestUser4', 'UBS', 'S')
        if u.data is not None:
            u.delete()


if __name__ == '__main__':
    unittest.main()



import unittest
import sys
import os
import time
import pandas as pd

from almapiwrapper.acquisitions import POLine, Vendor, Invoice, InvoiceLine, fetch_invoices
from almapiwrapper.inventory import Item
from almapiwrapper.record import JsonData, XmlData
from almapiwrapper import config_log

config_log("test.log")
if os.getcwd().endswith('test'):
    os.chdir('..')

class TestInvoice(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        pol_number = 'POL-UBS-2024-148772'
        pol = POLine(pol_number, zone='UBS', env='S')
        _ = pol.data
        if pol.error is True:
            data = JsonData(filepath='test/data/pol_test3.json')
            pol = POLine(data=data, zone='UBS', env='S').create()
            item = Item(barcode='UBS-S-1044370', zone='UBS', env='S')
            item.data.find('.//po_line').text = pol.pol_number
            item.update()

        invoice = Invoice(invoice_number='20000077919999', zone='UBS', env='S')
        _ = invoice.data
        if invoice.error is True:
            data = JsonData(filepath='test/data/invoice_test1.json')
            invoice = Invoice(data=data, zone='UBS', env='S').create()

        if 'invoice_lines' not in invoice.data or invoice.data['invoice_lines']['total_record_count'] == 0:
            data = JsonData(filepath='test/data/invoice_test1.json')
            invoice_line_data = data.content['invoice_lines']['invoice_line'][0]
            invoice_line = InvoiceLine(invoice_id=invoice.invoice_id,
                                       data=invoice_line_data,
                                       zone='UBS',
                                       env='S').create()


    def test_update(self):
        invoice =  Invoice(invoice_number='20000077919999', zone='UBS', env='S')
        initial_voucher_date = invoice.data['payment']['voucher_date']
        if initial_voucher_date == '2023-04-10Z':
            new_date = '2023-05-10Z'
        else:
            new_date = '2023-04-10Z'

        invoice.data['payment']['voucher_date'] = new_date

        invoice_updated = invoice.update()
        self.assertEqual(invoice_updated.data['payment']['voucher_date'],
                         new_date,
                         f'Voucher date should be {new_date}')
        self.assertFalse(invoice_updated.error, f'Invoice update error: {invoice_updated.error}')

    @unittest.skip
    def test_create(self):
        invoice =  Invoice(invoice_number='20000077919999', zone='UBS', env='S')
        invoice.invoice_number = '20000077919999_copy'
        new_invoice = Invoice(data=invoice.data, zone='UBS', env='S').create()

        self.assertEqual('PO-UBS-4828001_copy',
                         new_invoice.invoice_number,
                         f'Invoice number should be PO-UBS-4828001_copy')
        self.assertFalse(new_invoice.error, f'Invoice create error: {new_invoice.error}')

    def test_get_invoice_lines(self):
        invoice =  Invoice(invoice_number='20000077919999', zone='UBS', env='S')
        invoice_lines = invoice.get_invoice_lines()
        self.assertNotEqual(len(invoice_lines), 0, 'No invoice line found')

        # self.assertEqual(invoice_lines[0].invoice_line_id, '2000007791-1', 'Invoice line id should be 2000007791-1')


if __name__ == '__main__':
    unittest.main()



"""This module allows to get and update information about collections"""

from typing import Optional, Literal, List, ClassVar, Union
import logging
from ..record import Record, check_error, JsonData
import almapiwrapper.inventory as inventory


class Collection(Record):
    """
    Class representing a collection object

    Collections can be in NZ and in IZ. They contain list of bibliographic records.

    :ivar zone: zone of the record
    :ivar env: environment of the entity: 'P' for production and 'S' for sandbox
    :ivar data: :class:`almapiwrapper.record.JsonData` object or dict, useful to force update a record from a backup
    :ivar error: boolean indicating if an error occurred during the last operation
    :ivar error_msg: string containing the error message if an error occurred during the last operation
    :ivar pid: collection ID

    """

    api_base_url_bibs: ClassVar[str] = f'{Record.api_base_url}/bibs'

    def __init__(self,
                 pid: str,
                 zone: str,
                 env: Optional[Literal['P', 'S']] = 'P') -> None:
        """Construct a Collection record

        :param pid: collection ID
        :param zone: zone of the record
        :param env: environment of the entity: 'P' for production and 'S' for sandbox

        """
        self.pid = pid
        self._bibs = None
        self.error = False
        self.error_msg = None
        self._data = None
        self.area = 'Bibs'
        self.format = 'json'
        self.zone = zone
        self.env = env

    def __repr__(self) -> str:
        """Get a string representation of the object. Useful for logs.

        :return: string
        """
        return f"{self.__class__.__name__}('{self.pid}', '{self.zone}', '{self.env}')"

    def _fetch_data(self) -> Optional[JsonData]:
        """Fetch holding data via API. Store the data in the 'data' attribute.
        """
        r = self._api_call('get',
                           f'{self.api_base_url_bibs}/collections/{self.pid}',
                           headers=self._get_headers())

        if r.ok is True:
            logging.info(f'{repr(self)}: collection data available')
            return JsonData(r.json())
        else:
            self._handle_error(r, 'unable to fetch collection data')



    @check_error
    def save(self) -> 'Collection':
        """save(self) -> 'Collection'
        Save collection in a folder.

        Example: records/UBS_9963486250105504/hol_22314215780005504_01.xml

        :return: Collection
        """
        filepath = f'records/{self.zone}_collections/hol_{self.pid}.xml'
        self._save_from_path(filepath)
        return self

    def _fetch_bibs(self):
        """Fetch bibs of the collection via API. Store the data in the 'bibs' attribute.

        """

        self._bibs = []
        rec_count = None
        mms_ids = []
        while rec_count is None or len(mms_ids) < rec_count:
            r = self._api_call('get',
                               f'{self.api_base_url_bibs}/collections/{self.pid}/bibs',
                               params={'limit': '100', 'offset': str(len(mms_ids))},
                               headers=self._get_headers())
            if r.ok is False:
                self._handle_error(r, f'{repr(self)}: unable to fetch set members')
                return
            data = r.json()
            rec_count = data['total_record_count']
            if 'bib' in data:
                mms_ids += [rec['mms_id'] for rec in data['bib']]
                logging.info(f'{repr(self)}: {len(mms_ids)} / {rec_count} records fetched')
        if self.zone != 'NZ':
            self._bibs = [inventory.IzBib(mms_id, self.zone, self.env) for mms_id in mms_ids]
        else:
            self._bibs = [inventory.NzBib(mms_id, self.env) for mms_id in mms_ids]


    @property
    @check_error
    def bibs(self) -> Optional[List]:
        """bibs(self) -> Optional[List]
        Property of the collection returning the list containing bib records

        :return: List containing bib records
        """
        if self._bibs is None:
            self._fetch_bibs()

        return self._bibs

    def add_bib(self, bib: Union[inventory.IzBib, inventory.NzBib, str]) -> None:
        """add_bib(self, bib: Union[inventory.IzBib, inventory.NzBib]) -> None
        Add a bib to the collection

        :param bib: bib to add to the collection

        :return: None
        """

        if isinstance(bib, inventory.IzBib) or isinstance(bib, inventory.NzBib):
            mms_id = bib.mms_id
        else:
            mms_id = bib
        xml = f'<bib><mms_id>{mms_id}</mms_id></bib>'

        r = self._api_call('post',
                            f'{self.api_base_url_bibs}/collections/{self.pid}/bibs',
                            data=xml,
                            headers=self._get_headers(data_format='xml'))

        if r.ok is False:
            self._handle_error(r, f'{repr(self)}: unable to add bib {mms_id} to collection')
            return
        self._bibs = None
        logging.info(f'{repr(self)}: bib {mms_id} added to collection')

    def remove_bib(self, bib: Union[inventory.IzBib, inventory.NzBib, str]) -> None:
        """remove_bib(self, bib: Union[inventory.IzBib, inventory.NzBib]) -> None
        Remove a bib from the collection

        :param bib: bib to remove from the collection, can be either a NZ
            or an IZ bib or a mms_id

        :return: None
        """

        if isinstance(bib, inventory.IzBib) or isinstance(bib, inventory.NzBib):
            mms_id = bib.mms_id
        else:
            mms_id = bib

        r = self._api_call('delete',
                            f'{self.api_base_url_bibs}/collections/{self.pid}/bibs/{mms_id}',
                            headers=self._get_headers())

        if r.ok is False:
            self._handle_error(r, f'{repr(self)}: unable to remove bib {mms_id} from collection')
            return
        self._bibs = None
        logging.info(f'{repr(self)}: bib {mms_id} removed from collection')



import unittest
import sys
import os
import time
import pandas as pd

from almapiwrapper.acquisitions import POLine, Vendor
from almapiwrapper.record import JsonData, XmlData
from almapiwrapper import config_log

config_log("test.log")
if os.getcwd().endswith('test'):
    os.chdir('..')

class TestVendor(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        Vendor('UBS-A100-1021_copy', 'UBS', 'S').delete()

    def test_update(self):
        v =  Vendor('UBS-A100-1021', 'UBS', 'S')
        name = v.data['name']
        if name == 'IFF Institut für Finanzwirtschaft und Finanzrecht':
            v.data['name'] = 'IFF Institut'
        else:
            v.data['name'] = 'IFF Institut für Finanzwirtschaft und Finanzrecht'
        name = v.data['name']
        v.update()

        self.assertEqual(v.data['name'],
                         name,
                         f'Vendor name should be {name}')
        self.assertFalse(v.error, f'Vendor update error: {v.error}')

    def test_create(self):
        v =  Vendor('UBS-A100-1021', 'UBS', 'S')
        v.data['code'] = 'UBS-A100-1021_copy'
        for account in v.data['account']:
            del account['account_id']
        _ = Vendor(zone='UBS', env='S', data=v.data).create()
        v_copy = Vendor('UBS-A100-1021_copy', zone='UBS', env='S')

        self.assertEqual('UBS-A100-1021_copy',
                         v_copy.data['code'],
                         f'Vendor code should be UBS-A100-1021_copy')
        self.assertFalse(v.error, f'Vendor update error: {v.error}')

if __name__ == '__main__':
    unittest.main()



from ..record import Record, check_error, JsonData
from ..inventory import Item
from typing import Optional, Literal,ClassVar, Union, List
import almapiwrapper.acquisitions as acquisitionslib
from lxml import etree
import pandas as pd
import logging

class POLine(Record):
    """Class representing a POLine

    """

    api_base_url_acquisitions: ClassVar[str] = f'{Record.api_base_url}/acq/po-lines'

    def __init__(self,
                 pol_number: Optional[str] = None,
                 zone: Optional[str] = None,
                 env: Literal['P', 'S'] = 'P',
                 data: Optional[Union[dict, JsonData]] = None) -> None:
        """Constructor of POLine Object
        """

        super().__init__(zone, env)
        self.area = 'Acquisitions'
        self.format = 'json'
        self.pol_number = pol_number
        if data is not None:
            self._data = data if type(data) is JsonData else JsonData(data)
        elif pol_number is not None:
            pass
        else:
            self.error = True
            logging.error('Missing information to construct an POLine')

    def __repr__(self) -> str:
        """Get a string representation of the object. Useful for logs.
        :return: str
        """
        return f"{self.__class__.__name__}('{self.pol_number}', '{self.zone}', '{self.env}')"

    def _fetch_data(self) -> Optional[JsonData]:
        """Fetch the json data of the POLine

        :return: :class:`almapiwrapper.record.JsonData` if no error else None
        """

        r = self._api_call('get',
                     f'{self.api_base_url_acquisitions}/{self.pol_number}',
                           headers=self._get_headers())
        if r.ok is True:
            # Parse data
            json_data = JsonData(r.json())
            logging.debug(f"{self.__class__.__name__} data fetched")

            return json_data
        else:
            self._handle_error(r, 'unable to fetch POLine data')

    @check_error
    def update(self, update_inventory=True) -> 'acquisitionslib.POLine':
        """Update the POLine

        :return: POLine object
        """
        params = {'update_inventory': 'true'} if update_inventory is True else {'update_inventory': 'false'}
        r = self._api_call('put',
                           f'{self.api_base_url_acquisitions}/{self.pol_number}',
                           headers=self._get_headers(),
                           data=bytes(self),
                           params=params)

        if r.ok is True:
            self.data = JsonData(r.json())
            logging.info(f'{repr(self)}: POLine data updated')
        else:
            self._handle_error(r, 'unable to update POLine data')

        return self

    @check_error
    def save(self) -> 'acquisitionslib.POLine':
        """save(self) -> acquisitionslib.POLine
        Save a PO Line record in the 'records' folder

        When saved, a suffix is added to the file path with the version.
        Example: records/polines/pol_<IZ>_<set_id>_<version>.xml

        :return: object :class:`almapiwrapper.acquisitions.POLine`

        .. note::
            If the record encountered an error, this
            method will be skipped.
        """
        filepath = f'records/polines/pol_{self.zone}_{self.pol_number}.json'
        self._save_from_path(filepath)
        return self

    @check_error
    def create(self):
        r = self._api_call('post',
                           f'{self.api_base_url_acquisitions}',
                           headers=self._get_headers(),
                           data=bytes(self))

        if r.ok is True:
            self._data = JsonData(r.json())
            self.pol_number = self.data['number']
            logging.info(f'{repr(self)}: POLine data created: {self.pol_number}')
        else:
            self._handle_error(r, 'unable to create POLine')

        return self

    @check_error
    def get_items(self) -> List[Item]:
        """Get the items of the POLine

        :return: list of :class:`almapiwrapper.inventory.Item`
        """
        items = []
        for location in self.data['location']:
            for item in location['copy']:
                items.append(Item(barcode=item['barcode'],
                                  zone=self.zone,
                                  env=self.env))
        return items

    @check_error
    def get_vendor(self) -> Optional['acquisitionslib.Vendor']:
        """Get the vendor of the POLine

        :return: :class:`almapiwrapper.acquisitions.Vendor` or None if not available
        """
        if 'vendor' in self.data and 'value' in self.data['vendor']:
            return acquisitionslib.Vendor(vendor_code=self.data['vendor']['value'],
                                          zone=self.zone,
                                          env=self.env)
        else:
            return None

    @check_error
    def receive_item(self, item: Item, library: Optional[str] = None, department: Optional[str] = None, receive_date: Optional[str] = None) -> 'acquisitionslib.POLine':
        """Receive an item of the POLine

        :param item: :class:`almapiwrapper.inventory.Item`
        :param receive_date: str : date of the reception in format YYYY-MM-DDZ

        :return: POLine object
        """
        if item.error is True:
            logging.error(f'{repr(self)}: unable to receive item {repr(item)}, item has an error')
            return self
        params = {'op': 'receive'}
        if receive_date is not None:
            params['receive_date'] = receive_date
        if library is not None:
            params['department_library'] = library
        if department is not None:
            params['department'] = department
        headers = self._get_headers()
        headers['content-type'] = 'application/xml'

        r = self._api_call('post',
                           f'{self.api_base_url_acquisitions}/{self.pol_number}/items/{item.item_id}',
                           params=params,
                           headers=headers,
                           data='<item />')

        if r.ok is True:
            logging.info(f'{repr(self)}: Item {repr(item)} received.')
        else:
            self._handle_error(r, 'unable to receive item')

        return self
